"""
===========================
Plotting feature importance
===========================

A simple example showing how to compute and display
feature importances, it is also compared with the
feature importances obtained using random forests.

Feature importance is a measure of the effect of the features
on the outputs. For each feature, the values go from
0 to 1 where a higher the value means that the feature will have
a higher effect on the outputs.

Currently three criteria are supported : 'gcv', 'rss' and 'nb_subsets'.
See [1], section 12.3 for more information about the criteria.

.. [1] http://www.milbo.org/doc/earth-notes.pdf

"""

from __future__ import annotations

import matplotlib.pyplot as plt
import numpy
from sklearn.ensemble import RandomForestRegressor

from pyearth import Earth

# Create some fake data
numpy.random.seed(2)
m = 10000
n = 10

X = numpy.random.uniform(size=(m, n))
y = (
    10 * numpy.sin(numpy.pi * X[:, 0] * X[:, 1])
    + 20 * (X[:, 2] - 0.5) ** 2
    + 10 * X[:, 3]
    + 5 * X[:, 4]
    + numpy.random.uniform(size=m)
)
# Fit an Earth model
criteria = ("rss", "gcv", "nb_subsets")
model = Earth(
    max_degree=3,
    max_terms=10,
    minspan_alpha=0.5,
    feature_importance_type=criteria,
    verbose=True,
)
model.fit(X, y)
rf = RandomForestRegressor()
rf.fit(X, y)
# Print the model
print(model.trace())
print(model.summary())
print(model.summary_feature_importances(sort_by="gcv"))

# Plot the feature importances
importances = model.feature_importances_
importances["random_forest"] = rf.feature_importances_
criteria = criteria + ("random_forest",)
idx = 1

fig = plt.figure(figsize=(20, 10))
labels = ["$x_{}$".format(i) for i in range(n)]
for crit in criteria:
    plt.subplot(2, 2, idx)
    plt.bar(numpy.arange(len(labels)), importances[crit], align="center", color="red")
    plt.xticks(numpy.arange(len(labels)), labels)
    plt.title(crit)
    plt.ylabel("importances")
    idx += 1
title = (
    "$x_0,...x_9 \\sim \\mathcal{N}(0, 1)$\n$y= 10sin(\\pi x_{0}x_{1}) + 20(x_2 - 0.5)^2 + 10x_3 + 5x_4 + Unif(0, 1)$"
)
fig.suptitle(title, fontsize="x-large")
plt.show()



# distutils: language = c
# cython: cdivision = True
# cython: boundscheck = False
# cython: wraparound = False
# cython: profile = False

import numpy as np
from libc.math cimport sqrt, log

cdef FLOAT_t log2(FLOAT_t x):
    return log(x) / log(2.0)

cpdef apply_weights_2d(cnp.ndarray[FLOAT_t, ndim=2] B,
                       cnp.ndarray[FLOAT_t, ndim=1] weights):
    cdef INDEX_t i
    cdef INDEX_t j
    cdef INDEX_t m = B.shape[0]
    cdef INDEX_t n = B.shape[1]
    for i in range(m):
        for j in range(n):
            B[i, j] *= sqrt(weights[i])

cpdef apply_weights_slice(cnp.ndarray[FLOAT_t, ndim=2] B,
                          cnp.ndarray[FLOAT_t, ndim=1] weights,
                          INDEX_t column):
    cdef INDEX_t i
    cdef INDEX_t j
    cdef INDEX_t m = B.shape[0]
    cdef INDEX_t n = B.shape[1]
    for i in range(m):
        B[i, column] *= sqrt(weights[i])

cpdef apply_weights_1d(cnp.ndarray[FLOAT_t, ndim=1] y,
                       cnp.ndarray[FLOAT_t, ndim=1] weights):
    cdef INDEX_t i
    cdef INDEX_t m = y.shape[0]
    for i in range(m):
        y[i] *= sqrt(weights[i])

cpdef FLOAT_t gcv(FLOAT_t mse, FLOAT_t basis_size, FLOAT_t data_size,
                  FLOAT_t penalty):
    return mse * gcv_adjust(basis_size, data_size, penalty)

cpdef FLOAT_t gcv_adjust(FLOAT_t basis_size, FLOAT_t data_size,
                         FLOAT_t penalty):
    cdef FLOAT_t effective_parameters
    effective_parameters = basis_size + penalty * (basis_size - 1) / 2.0
    return 1.0 / ( ( (1.0 - (effective_parameters / data_size)) ** 2 ) )

cpdef str_pad(string, length):
    if len(string) >= length:
        return string[0:length]
    pad = length - len(string)
    return string + ' ' * pad

cpdef ascii_table(header, data, print_header=True, print_footer=True):
    '''
    header - list of strings representing the header row
    data - list of lists of strings representing data rows
    '''
    m = len(data)
    n = len(header)
    column_widths = [len(head) for head in header]
    for i, row in enumerate(data):
        for j, col in enumerate(row):
            if len(col) > column_widths[j]:
                column_widths[j] = len(col)

    for j in range(n):
        column_widths[j] += 1

    result = ''
    if print_header:
        for j, col_width in enumerate(column_widths):
            result += '-' * col_width + '-'
        result += '\n'
        for j, head in enumerate(header):
            result += str_pad(head, column_widths[j]) + ' '
        result += '\n'
        for j, col_width in enumerate(column_widths):
            result += '-' * col_width + '-'
#         result += '\n'
        result += '\n'
    for i, row in enumerate(data):
        if i > 0:
            result += '\n'
        for j, item in enumerate(row):
            result += str_pad(item, column_widths[j]) + ' '

    if print_footer:
        result += '\n'
        for j, col_width in enumerate(column_widths):
            result += '-' * col_width + '-'
    return result



from __future__ import annotations

import pickle

import numpy

from pyearth._basis import (
    ConstantBasisFunction,
    HingeBasisFunction,
    MissingnessBasisFunction,
)
from pyearth._types import BOOL

from . import assert_equal, assert_true
from .base import BaseContainer


class Container(BaseContainer):
    def __init__(self):
        super(Container, self).__init__()
        self.parent = ConstantBasisFunction()
        self.bf = MissingnessBasisFunction(self.parent, 1, True)
        self.child = HingeBasisFunction(self.bf, 1.0, 10, 1, False)


#
# def test_getters():
#     cnt = Container()
#     assert not cnt.bf.get_reverse()
#     assert cnt.bf.get_knot() == 1.0
#     assert cnt.bf.get_variable() == 1
#     assert cnt.bf.get_knot_idx() == 10
#     assert cnt.bf.get_parent() == cnt.parent


def test_apply():
    cnt = Container()
    m, _ = cnt.X.shape
    missing = numpy.zeros_like(cnt.X, dtype=BOOL)
    missing[1, 1] = True
    B = numpy.ones(shape=(m, 10))
    X = cnt.X.copy()
    X[1, 1] = None
    cnt.bf.apply(cnt.X, missing, B[:, 0])
    numpy.testing.assert_almost_equal(B[:, 0], 1 - missing[:, 1])
    cnt.child.apply(cnt.X, missing, B[:, 0])
    expected = (cnt.X[:, 1] - 1.0) * (cnt.X[:, 1] > 1.0)
    expected[1] = 0.0
    numpy.testing.assert_almost_equal(B[:, 0], expected)


# def test_apply_deriv():
#     cnt = Container()
#     m, _ = cnt.X.shape
#     missing = numpy.zeros_like(cnt.X, dtype=BOOL)
#     b = numpy.empty(shape=m)
#     j = numpy.empty(shape=m)
#     cnt.bf.apply_deriv(cnt.X, missing, b, j, 1)
#     numpy.testing.assert_almost_equal(
#         (cnt.X[:, 1] - 1.0) * (cnt.X[:, 1] > 1.0),
#         b
#     )
#     numpy.testing.assert_almost_equal(1.0 * (cnt.X[:, 1] > 1.0), j)


def test_degree():
    cnt = Container()
    assert_equal(cnt.bf.degree(), 1)


def test_pickle_compatibility():
    cnt = Container()
    bf_copy = pickle.loads(pickle.dumps(cnt.bf))
    assert_true(cnt.bf == bf_copy)


#
# def test_smoothed_version():
#     cnt = Container()
#     knot_dict = {cnt.bf: (.5, 1.5)}
#     translation = {cnt.parent: cnt.parent._smoothed_version(None, {}, {})}
#     smoothed = cnt.bf._smoothed_version(cnt.parent, knot_dict,
#                                         translation)
#
#     assert_true(type(smoothed) is SmoothedHingeBasisFunction)
#     assert_true(translation[cnt.parent] is smoothed.get_parent())
#     assert_equal(smoothed.get_knot_minus(), 0.5)
#     assert_equal(smoothed.get_knot_plus(), 1.5)
#     assert_equal(smoothed.get_knot(), cnt.bf.get_knot())
#     assert_equal(smoothed.get_variable(), cnt.bf.get_variable())



"""
=====================================================
Exporting a fitted Earth models as a sympy expression
=====================================================

A simple example returning a sympy expression describing the fit of a sine function computed by Earth.

"""

from __future__ import annotations

import numpy

from pyearth import Earth, export

# Create some fake data
numpy.random.seed(2)
m = 1000
n = 10
X = 10 * numpy.random.uniform(size=(m, n)) - 40
y = 100 * (numpy.sin((X[:, 6])) - 4.0) + 10 * numpy.random.normal(size=m)

# Fit an Earth model
model = Earth(max_degree=2, minspan_alpha=0.5, verbose=False)
model.fit(X, y)

print(model.summary())

# return sympy expression
print("Resulting sympy expression:")
print(export.export_sympy(model))



from __future__ import annotations

import os

import numpy


class BaseContainer(object):
    filename = os.path.join(os.path.dirname(__file__), "../test_data.csv")
    data = numpy.genfromtxt(filename, delimiter=",", skip_header=1)
    X = numpy.array(data[:, 0:5])
    y = numpy.array(data[:, 5])



cimport cython
from ._types cimport FLOAT_t, INT_t, INDEX_t, BOOL_t
from ._basis cimport BasisFunction
from ._qr cimport UpdatingQT

@cython.final
cdef class SingleWeightDependentData:
    cdef readonly UpdatingQT updating_qt
    cdef readonly FLOAT_t[:] w
    cdef readonly INDEX_t m
    cdef readonly INDEX_t k
    cdef readonly INDEX_t max_terms
    cdef readonly FLOAT_t[:, :] Q_t
    cdef readonly FLOAT_t total_weight
#     cpdef int update_from_basis_function(SingleWeightDependentData self, BasisFunction bf, FLOAT_t[:,:] X,
#                                          BOOL_t[:,:] missing) except *
    cpdef int update_from_array(SingleWeightDependentData self, FLOAT_t[:] b) except *
#     cpdef int _update(SingleWeightDependentData self, FLOAT_t zero_tol)
    cpdef downdate(SingleWeightDependentData self)
    cpdef reweight(SingleWeightDependentData self, FLOAT_t[:] w, FLOAT_t[:,:] B, INDEX_t k)

@cython.final
cdef class MultipleOutcomeDependentData:
    cdef list outcomes
    cdef list weights
    cpdef update_from_array(MultipleOutcomeDependentData self, FLOAT_t[:] b)
    cpdef downdate(MultipleOutcomeDependentData self)
    cpdef list sse(MultipleOutcomeDependentData self)
    cpdef FLOAT_t mse(MultipleOutcomeDependentData self)

@cython.final
cdef class SingleOutcomeDependentData:
    cdef readonly FLOAT_t[:] y
    cdef readonly SingleWeightDependentData weight
    cdef readonly FLOAT_t[:] theta
    cdef public FLOAT_t omega
    cdef public FLOAT_t sse_
    cdef public INDEX_t m
    cdef public INDEX_t k
    cdef public INDEX_t max_terms
    cdef public object householder
    cpdef FLOAT_t sse(SingleOutcomeDependentData self)
    cpdef int synchronize(SingleOutcomeDependentData self) except *
    cpdef int update(SingleOutcomeDependentData self) except *
    cpdef downdate(SingleOutcomeDependentData self)


@cython.final
cdef class PredictorDependentData:
    cdef readonly FLOAT_t[:] p
    cdef readonly FLOAT_t[:] x
    cdef readonly FLOAT_t[:] candidates
    cdef readonly INDEX_t[:] order

@cython.final
cdef class KnotSearchReadOnlyData:
    cdef readonly PredictorDependentData predictor
    cdef readonly MultipleOutcomeDependentData outcome

@cython.final
cdef class KnotSearchState:
    cdef public FLOAT_t alpha
    cdef public FLOAT_t beta
    cdef public FLOAT_t lambda_
    cdef public FLOAT_t mu
    cdef public FLOAT_t upsilon
    cdef public FLOAT_t phi
    cdef public FLOAT_t phi_next
    cdef public INDEX_t ord_idx
    cdef public INDEX_t idx
    cdef public FLOAT_t zeta_squared

@cython.final
cdef class KnotSearchWorkingData:
    cdef readonly FLOAT_t[:] gamma
    cdef readonly FLOAT_t[:] kappa
    cdef readonly FLOAT_t[:] delta_kappa
    cdef readonly FLOAT_t[:] chi
    cdef readonly FLOAT_t[:] psi
    cdef KnotSearchState state

@cython.final
cdef class KnotSearchData:
    cdef readonly KnotSearchReadOnlyData constant
    cdef readonly list workings
    cdef public INDEX_t q

cdef dot(FLOAT_t[:] x1, FLOAT_t[:] x2, INDEX_t q)
cdef w2dot(FLOAT_t[:] w, FLOAT_t[:] x1, FLOAT_t[:] x2, INDEX_t q)
cdef wdot(FLOAT_t[:] w, FLOAT_t[:] x1, FLOAT_t[:] x2, INDEX_t q)
cdef inline void fast_update(PredictorDependentData predictor, SingleOutcomeDependentData outcome,
                        KnotSearchWorkingData working, FLOAT_t[:] p, INDEX_t q, INDEX_t m ,INDEX_t r) except *
cpdef tuple knot_search(KnotSearchData data, FLOAT_t[:] candidates, FLOAT_t[:] p, INDEX_t q, INDEX_t m, INDEX_t r, INDEX_t n_outcomes, int verbose)



"""
Created on Feb 16, 2013

@author: jasonrudy
"""

from __future__ import annotations

import numpy

from pyearth._basis import (
    Basis,
    ConstantBasisFunction,
    HingeBasisFunction,
    LinearBasisFunction,
)
from pyearth._forward import ForwardPasser
from pyearth._types import BOOL

numpy.random.seed(0)
basis = Basis(10)
constant = ConstantBasisFunction()
basis.append(constant)
bf1 = HingeBasisFunction(constant, 0.1, 10, 1, False, "x1")
bf2 = HingeBasisFunction(constant, 0.1, 10, 1, True, "x1")
bf3 = LinearBasisFunction(bf1, 2, "x2")
basis.append(bf1)
basis.append(bf2)
basis.append(bf3)
X = numpy.random.normal(size=(100, 10))
missing = numpy.zeros_like(X).astype(BOOL)
B = numpy.empty(shape=(100, 4), dtype=numpy.float64)
basis.transform(X, missing, B)
beta = numpy.random.normal(size=4)
y = numpy.empty(shape=100, dtype=numpy.float64)
y[:] = numpy.dot(B, beta) + numpy.random.normal(size=100)
sample_weight = numpy.ones((X.shape[0], 1))


def test_run(snapshot):
    forwardPasser = ForwardPasser(X, missing, y[:, numpy.newaxis], sample_weight, max_terms=1000, penalty=1)

    forwardPasser.run()
    res = str(forwardPasser.get_basis()) + "\n" + str(forwardPasser.trace())
    assert res == snapshot



"""
===================================================================
Plotting a multicolumn regression problem that includes missingness
===================================================================

An example plotting a simultaneous fit of the sine and cosine functions.
There are two redundant predictors, each of which has independent and random
missingness.
"""

from __future__ import annotations

import matplotlib.pyplot as plt
import numpy

from pyearth import Earth

# Create some fake data
numpy.random.seed(2)
m = 10000
n = 10
X = 80 * numpy.random.uniform(size=(m, n)) - 40
X[:, 5] = X[:, 6] + numpy.random.normal(0, 0.1, m)
y1 = 100 * (numpy.sin((X[:, 5] + X[:, 6]) / 20) - 4.0) + 10 * numpy.random.normal(size=m)
y2 = 100 * (numpy.cos((X[:, 5] + X[:, 6]) / 20) - 4.0) + 10 * numpy.random.normal(size=m)
y = numpy.concatenate([y1[:, None], y2[:, None]], axis=1)
missing = numpy.random.binomial(1, 0.2, (m, n)).astype(bool)
X_full = X.copy()
X[missing] = None
idx5 = (1 - missing[:, 5]).astype(bool)
idx6 = (1 - missing[:, 6]).astype(bool)

# Fit an Earth model
model = Earth(
    max_degree=5,
    minspan_alpha=0.5,
    allow_missing=True,
    enable_pruning=True,
    thresh=0.001,
    smooth=True,
    verbose=True,
)
model.fit(X, y)

# Print the model
print(model.trace())
print(model.summary())

# Plot the model
y_hat = model.predict(X)
fig = plt.figure()

for j in [0, 1]:
    ax1 = fig.add_subplot(3, 4, 1 + 2 * j)
    ax1.plot(X_full[idx5, 5], y[idx5, j], "b.")
    ax1.plot(X_full[idx5, 5], y_hat[idx5, j], "r.")
    ax1.set_xlim(-40, 40)
    ax1.set_title("x5 present")
    ax1.set_xlabel("x5")
    ax1.set_ylabel("sin" if j == 0 else "cos")

    ax2 = fig.add_subplot(3, 4, 2 + 2 * j)
    ax2.plot(X_full[idx6, 6], y[idx6, j], "b.")
    ax2.plot(X_full[idx6, 6], y_hat[idx6, j], "r.")
    ax2.set_xlim(-40, 40)
    ax2.set_title("x6 present")
    ax2.set_xlabel("x6")
    ax2.set_ylabel("sin" if j == 0 else "cos")

    ax3 = fig.add_subplot(3, 4, 5 + 2 * j, sharex=ax1)
    ax3.plot(X_full[~idx6, 5], y[~idx6, j], "b.")
    ax3.plot(X_full[~idx6, 5], y_hat[~idx6, j], "r.")
    ax3.set_title("x6 missing")
    ax3.set_xlabel("x5")
    ax3.set_ylabel("sin" if j == 0 else "cos")

    ax4 = fig.add_subplot(3, 4, 6 + 2 * j, sharex=ax2)
    ax4.plot(X_full[~idx5, 6], y[~idx5, j], "b.")
    ax4.plot(X_full[~idx5, 6], y_hat[~idx5, j], "r.")
    ax4.set_title("x5 missing")
    ax4.set_xlabel("x6")
    ax4.set_ylabel("sin" if j == 0 else "cos")

    ax5 = fig.add_subplot(3, 4, 9 + 2 * j, sharex=ax1)
    ax5.plot(X_full[(~idx6) & (~idx5), 5], y[(~idx6) & (~idx5), j], "b.")
    ax5.plot(X_full[(~idx6) & (~idx5), 5], y_hat[(~idx6) & (~idx5), j], "r.")
    ax5.set_title("both missing")
    ax5.set_xlabel("x5")
    ax5.set_ylabel("sin" if j == 0 else "cos")

    ax6 = fig.add_subplot(3, 4, 10 + 2 * j, sharex=ax2)
    ax6.plot(X_full[(~idx6) & (~idx5), 6], y[(~idx6) & (~idx5), j], "b.")
    ax6.plot(X_full[(~idx6) & (~idx5), 6], y_hat[(~idx6) & (~idx5), j], "r.")
    ax6.set_title("both missing")
    ax6.set_xlabel("x6")
    ax6.set_ylabel("sin" if j == 0 else "cos")

fig.tight_layout()
plt.show()



# distutils: language = c
# cython: cdivision = True
# cython: boundscheck = False
# cython: wraparound = False
# cython: profile = False
# cython: language-level = 3str

from ._util cimport log2, apply_weights_2d
from libc.math cimport log
from libc.math cimport abs
cimport cython
cimport numpy as cnp
from ._types cimport FLOAT_t
# cdef FLOAT_t ZERO_TOL = 1e-16
cdef cnp.float_t ZERO_TOL = 1e-16
from ._types import FLOAT
import numpy as np
import sys

max_int = sys.maxsize

cdef class BasisFunction:

    def __cinit__(BasisFunction self):
        self.pruned = False
        self.children = []
        self.prunable = True
        self.child_map = {}
        self.splittable = True

    def __hash__(BasisFunction self):
        return id(self) % max_int # avoid "OverflowError Python
                                     # int too large to convert to C long"

    cpdef smooth(BasisFunction self, dict knot_dict, dict translation):
        '''
        Modifies translation in place.
        '''
        cdef INDEX_t i, n = len(self.children)
        translation[self] = self._smoothed_version(self.get_parent(), knot_dict,
                                                   translation)
        for i in range(n):
            self.children[i].smooth(knot_dict, translation)

    def __reduce__(BasisFunction self):
        return (self.__class__, (), self._getstate())

    def _get_root(BasisFunction self):
        return self.parent._get_root()

    def _getstate(BasisFunction self):
        result = {'pruned': self.pruned,
                  'children': self.children,
                  'prunable': self.prunable,
                  'child_map': self.child_map,
                  'splittable': self.splittable}
        result.update(self._get_parent_state())
        return result

    def _get_parent_state(BasisFunction self):
        return {'parent': self.parent}

    def _set_parent_state(BasisFunction self, state):
        self.parent = state['parent']

    def __setstate__(BasisFunction self, state):
        self.pruned = state['pruned']
        self.children = state['children']
        self.prunable = state['prunable']
        self.child_map = state['child_map']
        self.splittable = state['splittable']
        self._set_parent_state(state)

    def _eq(BasisFunction self, other):
        if self.__class__ is not other.__class__:
            return False
        self_state = (self._getstate(), self.__reduce__()[1])
        other_state = (other._getstate(), other.__reduce__()[1])
        del self_state[0]['children']
        del self_state[0]['child_map']
        del other_state[0]['children']
        del other_state[0]['child_map']
        return self_state == other_state

    def __richcmp__(BasisFunction self, other, method):
        if method == 2:
            return self._eq(other)
        elif method == 3:
            return not self._eq(other)
        else:
            return NotImplemented

    cpdef bint has_knot(BasisFunction self):
        return False

    cpdef bint is_prunable(BasisFunction self):
        return self.prunable

    cpdef bint is_pruned(BasisFunction self):
        return self.pruned

    cpdef bint is_splittable(BasisFunction self):
        return self.splittable

    cpdef bint make_splittable(BasisFunction self):
        self.splittable = True

    cpdef bint make_unsplittable(BasisFunction self):
        self.splittable = False

    cpdef list get_children(BasisFunction self):
        return self.children

    cpdef BasisFunction get_coverage(BasisFunction self, INDEX_t variable):
        cdef BasisFunction child
        for child in self.get_children():
            if child.covered(variable):
                return child
        return None

    cpdef bool has_linear(BasisFunction self, INDEX_t variable):
        cdef BasisFunction child  # @DuplicatedSignature
        for child in self.get_children():
            if child.linear_in(variable):
                return True
        return False

    cpdef bool linear_in(BasisFunction self, INDEX_t variable):
        return False

    cpdef _set_parent(BasisFunction self, BasisFunction parent):
        '''Calls _add_child.'''
        self.parent = parent
        self.parent._add_child(self)

    cpdef _add_child(BasisFunction self, BasisFunction child):
        '''Called by _set_parent.'''
        cdef INDEX_t n = len(self.children)
        self.children.append(child)
        cdef int var = child.get_variable()
        if var in self.child_map:
            self.child_map[var].append(n)
        else:
            self.child_map[var] = [n]

    cpdef BasisFunction get_parent(BasisFunction self):
        return self.parent

    cpdef prune(BasisFunction self):
        self.pruned = True

    cpdef unprune(BasisFunction self):
        self.pruned = False

    cpdef knots(BasisFunction self, INDEX_t variable):

        cdef list children
        cdef BasisFunction child
        if variable in self.child_map:
            children = self.child_map[variable]
        else:
            return []
        cdef INDEX_t n = len(children)
        cdef INDEX_t i
        cdef list result = []
        cdef int idx
        for i in range(n):
            idx = children[i]
            child = self.get_children()[idx]
            if child.has_knot():
                result.append(child.get_knot_idx())
        return result

    cpdef INDEX_t effective_degree(BasisFunction self):
        cdef dict data_dict = {}
        cdef dict missing_dict = {}
        self._effective_degree(data_dict, missing_dict)
        cdef INDEX_t k, v
        for k, v in missing_dict.items():
            if k in data_dict:
                data_dict[k] += missing_dict[k] - 1
            else:
                data_dict[k] = missing_dict[k]
        return sum(data_dict.values())

    cpdef apply(BasisFunction self, cnp.ndarray[FLOAT_t, ndim=2] X,
                cnp.ndarray[BOOL_t, ndim=2] missing,
                cnp.ndarray[FLOAT_t, ndim=1] b, bint recurse=True):
        '''
        X - Data matrix
        b - parent vector
        recurse - If False, assume b already contains the result of
        the parent function.  Otherwise, recurse to compute parent function.
        '''

    cpdef cnp.ndarray[INT_t, ndim = 1] valid_knots(BasisFunction self,
        cnp.ndarray[FLOAT_t, ndim=1] values,
        cnp.ndarray[FLOAT_t, ndim=1] variable, int variable_idx,
        INDEX_t check_every,
        int endspan, int minspan,
        FLOAT_t minspan_alpha, INDEX_t n,
        cnp.ndarray[INT_t, ndim=1] workspace):
        '''
        values - The unsorted values of self in the data set
        variable - The sorted values of variable in the data set
        variable_idx - The index of the variable in the data set
        workspace - An m-vector (where m is the number of samples) used
                    internally
        '''
        cdef INDEX_t i
        cdef INDEX_t j
        cdef INDEX_t k
        cdef INDEX_t m = values.shape[0]
        cdef FLOAT_t float_tmp
        cdef INT_t int_tmp
        cdef INDEX_t count
        cdef int minspan_
        cdef cnp.ndarray[INT_t, ndim = 1] result
        cdef INDEX_t num_used
        cdef INDEX_t prev
        cdef INDEX_t start
        cdef int idx
        cdef int last_idx
        cdef FLOAT_t first_var_value = variable[m - 1]
        cdef FLOAT_t last_var_value = variable[m - 1]

        # Calculate the used knots
        cdef list used_knots = self.knots(variable_idx)
        used_knots.sort()

        # Initialize workspace to 1 where value is nonzero
        # Also, find first_var_value as the maximum variable
        # where value is nonzero and last_var_value to the
        # minimum variable where value is nonzero
        count = 0
        for i in range(m):
            if abs(values[i]) > ZERO_TOL:
                workspace[i] = 1
                count += 1
                if variable[i] >= first_var_value:
                    first_var_value = variable[i]
                last_var_value = variable[i]
            else:
                workspace[i] = 0

        # Calculate minspan
        if minspan < 0:
            minspan_ = <int > (-log2(-(1.0 / (n * count)) *
                                log(1.0 - minspan_alpha)) / 2.5)
        else:
            minspan_ = minspan

        # Take out the used points and apply minspan
        num_used = len(used_knots)
        prev = 0
        last_idx = -1
        for i in range(num_used):
            idx = used_knots[i]
            if last_idx == idx:
                continue
            workspace[idx] = 0
            j = idx
            k = 0
            while j > prev + 1 and k < minspan_:
                if workspace[j - 1]:
                    workspace[j - 1] = False
                    k += 1
                j -= 1
            j = idx + 1
            k = 0
            while j < m and k < minspan_:
                if workspace[j]:
                    workspace[j] = False
                    k += 1
                j += 1
            prev = idx
            last_idx = idx

        # Apply endspan
        i = 0
        j = 0
        while i < endspan:
            if workspace[j]:
                workspace[j] = 0
                i += 1
            j += 1
            if j == m:
                break
        i = 0
        j = m - 1
        while i < endspan:
            if workspace[j]:
                workspace[j] = 0
                i += 1
            if j == 0:
                break
            j -= 1

        # Implement check_every
        int_tmp = 0
        count = 0
        for i in range(m):
            if workspace[i]:
                if (int_tmp % check_every) != 0:
                    workspace[i] = 0
                else:
                    count += 1
                int_tmp += 1
            else:
                int_tmp = 0

        # Make sure the greatest value is not a candidate (this can happen if
        # the first endspan+1 values are the same)
        for i in range(m):
            if workspace[i]:
                if variable[i] == first_var_value:
                    workspace[i] = 0
                    count -= 1
                else:
                    break

        # Also make sure the least value is not a candidate
        for i in range(m):
            if workspace[m - i - 1]:
                if variable[m - i - 1] == last_var_value:
                    workspace[m - i - 1] = 0
                    count -= 1
                else:
                    break

        # Create result array and return
        result = np.empty(shape=count, dtype=int)
        j = 0
        for i in range(m):
            if workspace[i]:
                result[j] = i
                j += 1

        return result

    def func_factory(HingeBasisFunction self, coef):
        return eval(self.func_string_factory(coef))

cdef class PicklePlaceHolderBasisFunction(BasisFunction):
    '''This is a place holder for unpickling the basis function tree.'''

pickle_place_holder = PicklePlaceHolderBasisFunction()

cdef class RootBasisFunction(BasisFunction):
    def __init__(RootBasisFunction self):  # @DuplicatedSignature
        self.prunable = False

    cpdef bint covered(RootBasisFunction self, INDEX_t variable):
        '''
        Is this an covered parent for variable? (If not, a covering
        MissingnessBasisFunction must be added before the variable can
        be used).
        '''
        return False

    cpdef bint eligible(RootBasisFunction self, INDEX_t variable):
        '''
        Is this an eligible parent for variable?
        '''
        return True

    def copy(RootBasisFunction self):
        return self.__class__()

    def _get_root(RootBasisFunction self):  # @DuplicatedSignature
        return self

    def _get_parent_state(RootBasisFunction self):  # @DuplicatedSignature
        return {}

    def _set_parent_state(RootBasisFunction self, state): # @DuplicatedSignature
        pass

    cpdef set variables(RootBasisFunction self):
        return set()

    cpdef _smoothed_version(RootBasisFunction self, BasisFunction parent,
                            dict knot_dict, dict translation):
        result = self.__class__()
        if self.is_pruned():
            result.prune()
        return result

    cpdef INDEX_t degree(RootBasisFunction self):
        return 0

    cpdef _effective_degree(RootBasisFunction self, dict data_dict, dict missing_dict):
        pass

    cpdef _set_parent(RootBasisFunction self, BasisFunction parent):
        raise NotImplementedError

    cpdef BasisFunction get_parent(RootBasisFunction self):
        return None

    cpdef apply(RootBasisFunction self, cnp.ndarray[FLOAT_t, ndim=2] X,
                cnp.ndarray[BOOL_t, ndim=2] missing,
                cnp.ndarray[FLOAT_t, ndim=1] b, bint recurse=False):
        '''
        X - Data matrix
        missing - missingness matrix
        b - parent vector
        recurse - The ZeroBasisFunction is an alternative RootBasisFunction used
                  for computing derivatives.
        It is the derivative of the ConstantBasisFunction.
        '''
        b[:] = self.eval()

    cpdef apply_deriv(RootBasisFunction self, cnp.ndarray[FLOAT_t, ndim=2] X,
                      cnp.ndarray[BOOL_t, ndim=2] missing,
                      cnp.ndarray[FLOAT_t, ndim=1] b,
                      cnp.ndarray[FLOAT_t, ndim=1] j, INDEX_t var):
        '''
        X - Data matrix
        missing - missingness matrix
        b - holds the value of the basis function
        j - holds the value of the derivative
        '''
        b[:] = self.eval()
        j[:] = self.eval_deriv()

@cython.final
cdef class ConstantBasisFunction(RootBasisFunction):

    cpdef inline FLOAT_t eval(ConstantBasisFunction self):
        return <FLOAT_t> 1.0

    cpdef inline FLOAT_t eval_deriv(ConstantBasisFunction self):
        return <FLOAT_t> 0.0

    def __str__(ConstantBasisFunction self):
        return '(Intercept)'

    def func_string_factory(ConstantBasisFunction self, coef):
        if coef is not None:
            return "lambda x: {:s}".format(repr(coef))
        else:
            return ''

cdef class VariableBasisFunction(BasisFunction):
    cpdef INDEX_t degree(VariableBasisFunction self):
        return self.parent.degree() + 1

    cpdef set variables(VariableBasisFunction self):
        cdef set result = self.parent.variables()
        result.add(self.get_variable())
        return result

    cpdef INDEX_t get_variable(VariableBasisFunction self):
        return self.variable

cdef class DataVariableBasisFunction(VariableBasisFunction):
    cpdef _effective_degree(DataVariableBasisFunction self, dict data_dict, dict missing_dict):
        try:
            data_dict[self.variable] += 1
        except:
            data_dict[self.variable] = 1
        self.parent._effective_degree(data_dict, missing_dict)

    cpdef bint covered(DataVariableBasisFunction self, INDEX_t variable):
        '''
        Is this an covered parent for variable? (If not, a covering
        MissingnessBasisFunction must be added before the variable can
        be used).
        '''
        return False or self.parent.covered(variable)

    cpdef bint eligible(DataVariableBasisFunction self, INDEX_t variable):
        '''
        Is this an eligible parent for variable?
        '''
        return True and self.parent.eligible(variable)

    cpdef apply(DataVariableBasisFunction self, cnp.ndarray[FLOAT_t, ndim=2] X,
                cnp.ndarray[BOOL_t, ndim=2] missing,
                cnp.ndarray[FLOAT_t, ndim=1] b, bint recurse=True):
        '''
        X - Data matrix
        missing - missingness matrix
        b - parent vector
        recurse - If False, assume b already contains the result of the parent
                  function.  Otherwise, recurse to compute parent function.
        '''
        cdef INDEX_t i # @DuplicatedSignature
        cdef INDEX_t m = len(b)  # @DuplicatedSignature
        cdef cnp.ndarray[FLOAT_t, ndim=1] val
        if recurse:
            self.parent.apply(X, missing, b, recurse=True)
        val = np.zeros(X.shape[0], dtype=FLOAT)
        here =  missing[:, self.variable] == 0
        val[here] = self.eval(X[here, self.variable])
        for i in range(m):
            if not missing[i, self.variable]:
                b[i] *= val[i]

    cpdef apply_deriv(DataVariableBasisFunction self,
                      cnp.ndarray[FLOAT_t, ndim=2] X,
                      cnp.ndarray[BOOL_t, ndim=2] missing,
                      cnp.ndarray[FLOAT_t, ndim=1] b,
                      cnp.ndarray[FLOAT_t, ndim=1] j, INDEX_t var):
        '''
        X - Data matrix
        missing - missingness matrix
        j - result vector
        '''
        cdef INDEX_t i, this_var = self.get_variable()  # @DuplicatedSignature
        cdef INDEX_t m = len(b)  # @DuplicatedSignature
        cdef FLOAT_t x
        self.parent.apply_deriv(X, missing, b, j, var)
        here =  missing[:, self.variable] == 0
        this_val = np.zeros(X.shape[0], dtype=FLOAT)
        this_deriv = np.zeros(X.shape[0], dtype=FLOAT)
        this_val[here] = self.eval(X[here,this_var])
        this_deriv[here] = self.eval_deriv(X[here,this_var])
        for i in range(m):
            if missing[i, this_var]:
                continue
            x = X[i,this_var]
            j[i] = j[i]*this_val[i]
            if this_var == var:
                j[i] += b[i]*this_deriv[i]
            b[i] *= this_val[i]

@cython.final
cdef class MissingnessBasisFunction(VariableBasisFunction):
    def __init__(MissingnessBasisFunction self, BasisFunction parent,
                 INDEX_t variable, bint complement,
                 label=None):
        self.variable = variable
        self.complement = complement
        self.label = label if label is not None else 'x' + str(variable)
        self._set_parent(parent)

    cpdef _effective_degree(MissingnessBasisFunction self, dict data_dict, dict missing_dict):
        try:
            missing_dict[self.variable] += 1
        except:
            missing_dict[self.variable] = 1
        self.parent._effective_degree(data_dict, missing_dict)

    cpdef bint covered(MissingnessBasisFunction self, INDEX_t variable):
        '''
        Is this an covered parent for variable? (If not, a covering
        MissingnessBasisFunction must be added before the variable can
        be used).
        '''
        if self.complement and (variable == self.variable):
            return True
        else:
            return self.parent.covered(variable) or False

    cpdef bint eligible(MissingnessBasisFunction self, INDEX_t variable):
        '''
        Is this an eligible parent for variable?
        '''
        if (not self.complement) and (variable == self.variable):
            return False
        else:
            return self.parent.eligible(variable) and True

    cpdef apply(MissingnessBasisFunction self, cnp.ndarray[FLOAT_t, ndim=2] X,
                cnp.ndarray[BOOL_t, ndim=2] missing,
                cnp.ndarray[FLOAT_t, ndim=1] b, bint recurse=True):
        '''
        X - Data matrix
        b - parent vector
        recurse - If False, assume b already contains the result of the parent
                  function.  Otherwise, recurse to compute parent function.
        '''
        if recurse:
            self.parent.apply(X, missing, b, recurse=True)
        if self.complement:
            b *= (1 - missing[:, self.variable])
        else:
            b *= missing[:, self.variable]

    cpdef apply_deriv(MissingnessBasisFunction self,
                      cnp.ndarray[FLOAT_t, ndim=2] X,
                      cnp.ndarray[BOOL_t, ndim=2] missing,
                      cnp.ndarray[FLOAT_t, ndim=1] b,
                      cnp.ndarray[FLOAT_t, ndim=1] j, INDEX_t var):
        '''
        X - Data matrix
        j - result vector
        '''
        cdef INDEX_t i, this_var = self.get_variable()  # @DuplicatedSignature
        cdef INDEX_t m = len(b)  # @DuplicatedSignature
        cdef FLOAT_t x
        self.parent.apply_deriv(X, missing, b, j, var)
        if self.complement:
            this_val = 1.0 - missing[:,this_var]
        else:
            this_val = 1.0 - missing[:,this_var]
        this_deriv = 1.0
        for i in range(m):
            x = X[i,this_var]
            j[i] *= this_val[i]
            b[i] *= this_val[i]

    cpdef _smoothed_version(MissingnessBasisFunction self, BasisFunction parent,
                            dict knot_dict, dict translation):
        result = MissingnessBasisFunction(translation[parent], self.variable,
                                          self.complement, self.label)
        if self.is_pruned():
            result.prune()
        return result

    def __reduce__(MissingnessBasisFunction self):
        return (self.__class__,
                (pickle_place_holder, self.variable, self.complement,
                self.label),
                self._getstate())

    def __str__(MissingnessBasisFunction self):
        if self.complement:
            result = 'present(%s)' % self.label
        else:
            result = 'missing(%s)' % self.label
        parent = (str(self.parent)
                  if not self.parent.__class__ is ConstantBasisFunction
                  else '')
        if parent != '':
            result += '*%s' % (str(self.parent),)
        return result

    def func_string_factory(MissingnessBasisFunction self, coef):
        parent = self.parent.func_string_factory(None)
        parent = ' * ' + parent if parent else ''
        if self.complement:
            result = "(x[{:d}] is not None){:s}".format(
                self.variable,
                parent)
        else:
            result = "(x[{:d}] is None){:s}".format(
                self.variable,
                parent)
        if coef is not None:
            result = 'lambda x: {:s} * {:s}'.format(str(coef), result)
        return result

cdef class HingeBasisFunctionBase(DataVariableBasisFunction):
    cpdef bint has_knot(HingeBasisFunctionBase self):
        return True

    cpdef FLOAT_t get_knot(HingeBasisFunctionBase self):
        return self.knot

    cpdef bint get_reverse(HingeBasisFunctionBase self):
        return self.reverse

    cpdef INDEX_t get_knot_idx(HingeBasisFunctionBase self):
        return self.knot_idx

@cython.final
cdef class SmoothedHingeBasisFunction(HingeBasisFunctionBase):

    def __init__(SmoothedHingeBasisFunction self, BasisFunction parent,
                 FLOAT_t knot, FLOAT_t knot_minus,
                 FLOAT_t knot_plus, INDEX_t knot_idx,
                 INDEX_t variable, bint reverse,
                 label=None):
        self.knot = knot
        self.knot_minus= knot_minus
        self.knot_plus = knot_plus
        self.knot_idx = knot_idx
        self.variable = variable
        self.reverse = reverse
        self.label = label if label is not None else 'x' + str(variable)
        self._set_parent(parent)
        self._init_p_r()

    cpdef get_knot_minus(SmoothedHingeBasisFunction self):
        return self.knot_minus

    cpdef get_knot_plus(SmoothedHingeBasisFunction self):
        return self.knot_plus

    cpdef _smoothed_version(SmoothedHingeBasisFunction self,
                            BasisFunction parent, dict knot_dict,
                            dict translation):
        result = SmoothedHingeBasisFunction(translation[parent], self.knot,
                                            self.knot_minus, self.knot_plus,
                                            self.knot_idx, self.variable,
                                            self.reverse, self.label)
        if self.is_pruned():
            result.prune()
        return result

    cpdef _init_p_r(SmoothedHingeBasisFunction self):
        # See Friedman, 1991, eq (35)
        cdef FLOAT_t p_denom = self.knot_plus - self.knot_minus
        cdef FLOAT_t r_denom = p_denom
        p_denom *= p_denom
        r_denom *= p_denom
        if not self.reverse:
            self.p = (2*self.knot_plus + self.knot_minus - 3*self.knot) / p_denom
            self.r = (2*self.knot - self.knot_plus - self.knot_minus) / r_denom
        else:
            self.p = (3*self.knot - 2*self.knot_minus - self.knot_plus) / p_denom
            self.r = -1*(self.knot_minus + self.knot_plus - 2*self.knot) / r_denom

    cpdef get_p(SmoothedHingeBasisFunction self):
        return self.p

    cpdef get_r(SmoothedHingeBasisFunction self):
        return self.r

    def __str__(SmoothedHingeBasisFunction self):  # @DuplicatedSignature
        result = ''
        if self.variable is not None:
            if not self.reverse:
                result = 'C(%s|s=+1,%G,%G,%G)' % (self.label, self.knot_minus,
                                                  self.knot, self.knot_plus)
            else:
                result = 'C(%s|s=-1,%G,%G,%G)' % (self.label, self.knot_minus,
                                                  self.knot, self.knot_plus)
        parent = (str(self.parent)
                  if not self.parent.__class__ is ConstantBasisFunction
                  else '')
        if parent != '':
            result += '*%s' % (str(self.parent),)
        return result

    def __reduce__(SmoothedHingeBasisFunction self):  # @DuplicatedSignature
        return (self.__class__,
                (pickle_place_holder, self.knot,
                 self.knot_minus, self.knot_plus,
                 self.knot_idx, self.variable, self.reverse, self.label),
                self._getstate())

    def eval(SmoothedHingeBasisFunction self, x):
        # See Friedman, 1991, eq (34)
        if not self.reverse:
            tmp2 = x - self.knot_minus
            return np.where(x <= self.knot_minus, 0.0,
                np.where((self.knot_minus < x) & (x < self.knot_plus),
                    self.p*tmp2**2 + self.r*tmp2**3, x - self.knot))
        else:
            tmp2 = x - self.knot_plus
            return np.where(x <= self.knot_minus, self.knot - x,
                np.where((self.knot_minus < x) & (x < self.knot_plus),
                    self.p*tmp2**2 + self.r*tmp2**3, 0.0))

    def eval_deriv(SmoothedHingeBasisFunction self, x):
        # See Friedman, 1991, eq (34)
        if not self.reverse:
            tmp2 = x - self.knot_minus
            return np.where(x <= self.knot_minus, 0.0,
                np.where((self.knot_minus < x) & (x < self.knot_plus),
                    2.0*self.p*tmp2 + 3.0*self.r*tmp2**2, 1.0))
        else:
            tmp2 = x - self.knot_plus
            return np.where(x <= self.knot_minus, -1.0,
                np.where((self.knot_minus < x) & (x < self.knot_plus),
                    2.0*self.p*tmp2 + 3.0*self.r*tmp2**2, 0.0))

    def func_string_factory(SmoothedHingeBasisFunction self, coef):
        parent = self.parent.func_string_factory(None)
        parent = ' * ' + parent if parent else ''
        args = {"p" : self.p,
                "r": self.r,
                "t_minus": self.knot_minus,
                "t_plus": self.knot_plus,
                "t": self.knot,
                "idx": self.variable,
                "parent": parent}
        if not self.reverse:
            result = "(0 if x[{idx}] <= {t_minus} else (x[{idx}] - {t}) if x[{idx}] >= {t_plus} else ({p} * (x[{idx}] - {t_minus}) ** 2 + {r} * (x[{idx}] - {t_minus}) ** 3)){parent}".format(**args)
        else:
            result = "(-(x[{idx}] - {t}) if x[{idx}] <= {t_minus} else 0 if x[{idx}] >= {t_plus} else ({p} * (x[{idx}] - {t_plus}) ** 2 + {r} * (x[{idx}] - {t_plus}) ** 3)){parent}".format(**args)

        if coef is not None:
            result = 'lambda x: {:s} * {:s}'.format(str(coef), result)
        return result

@cython.final
cdef class HingeBasisFunction(HingeBasisFunctionBase):

    def __init__(HingeBasisFunction self, BasisFunction parent, FLOAT_t knot,
                 INDEX_t knot_idx, INDEX_t variable, bint reverse, label=None):
        self.knot = knot
        self.knot_idx = knot_idx
        self.variable = variable
        self.reverse = reverse
        self.label = label if label is not None else 'x' + str(variable)
        self._set_parent(parent)

    cpdef _smoothed_version(HingeBasisFunction self, BasisFunction parent,
                            dict knot_dict, dict translation):
        knot_minus, knot_plus = knot_dict[self]
        result = SmoothedHingeBasisFunction(translation[parent], self.knot,
                                            knot_minus, knot_plus,
                                            self.knot_idx, self.variable,
                                            self.reverse, self.label)
        if self.is_pruned():
            result.prune()
        return result

    def __reduce__(HingeBasisFunction self):
        return (self.__class__,
                ( pickle_place_holder, self.knot, self.knot_idx,
                  self.variable, self.reverse, self.label),
                self._getstate())

    def __str__(HingeBasisFunction self):
        result = ''
        if self.variable is not None:
            if not self.reverse:
                if self.knot >= 0:
                    result = 'h(%s-%G)' % (self.label, self.knot)
                else:
                    result = 'h(%s+%G)' % (self.label, -self.knot)
            else:
                result = 'h(%G-%s)' % (self.knot, self.label)
        parent = (str(self.parent)
                  if not self.parent.__class__ is ConstantBasisFunction
                  else '')
        if parent != '':
            result += '*%s' % (str(self.parent),)
        return result

    def eval(HingeBasisFunction self, x):
        if self.reverse:
            return np.where(x  > self.knot, 0.0, self.knot - x)
        else:
            return np.where(x <= self.knot, 0.0, x - self.knot)

    def eval_deriv(HingeBasisFunction self, x):
        if self.reverse:
            return np.where(x  > self.knot, 0.0, -1.0)
        else:
            return np.where(x <= self.knot, 0.0,  1.0)

    def func_string_factory(HingeBasisFunction self, coef):
        parent = self.parent.func_string_factory(None)
        parent = ' * ' + parent if parent else ''
        if self.reverse:
            result = "max(0, {:s} - x[{:d}]){:s}".format(
                str(self.knot),
                self.variable,
                parent)
        else:
            result = "max(0, x[{:d}] - {:s}){:s}".format(
                self.variable,
                str(self.knot),
                parent)
        if coef is not None:
            result = 'lambda x: {:s} * {:s}'.format(str(coef), result)
        return result

@cython.final
cdef class LinearBasisFunction(DataVariableBasisFunction):
    #@DuplicatedSignature
    def __init__(LinearBasisFunction self, BasisFunction parent,
                 INDEX_t variable, label=None):
        self.variable = variable
        self.label = label if label is not None else 'x' + str(variable)
        self._set_parent(parent)

    cpdef bool linear_in(LinearBasisFunction self, INDEX_t variable):
        return variable == self.variable

    cpdef _smoothed_version(LinearBasisFunction self, BasisFunction parent,
                            dict knot_dict, dict translation):
        result = LinearBasisFunction(translation[parent], self.variable,
                                     self.label)
        if self.is_pruned():
            result.prune()
        return result

    def __reduce__(LinearBasisFunction self):
        return (self.__class__,
                (pickle_place_holder, self.variable, self.label),
                self._getstate())

    def __str__(LinearBasisFunction self):
        result = self.label
        if not self.parent.__class__ is ConstantBasisFunction:
            parent = str(self.parent)
            result += '*' + parent
        return result

    def eval(LinearBasisFunction self, x):
        return x

    def eval_deriv(LinearBasisFunction self, x):
        return np.ones(len(x))

    def func_string_factory(LinearBasisFunction self, coef):
        parent = self.parent.func_string_factory(None)
        parent = ' * ' + parent if parent else ''
        result = "x[{:d}]{:s}".format(
            self.variable,
            parent)
        if coef is not None:
            result = 'lambda x: {:s} * {:s}'.format(str(coef), result)
        return result

cdef class Basis:
    '''A container that provides functionality related to a set of
    BasisFunctions with a common ConstantBasisFunction ancestor.
    Retains the order in which BasisFunctions are added.'''

    def __init__(Basis self, num_variables):  # @DuplicatedSignature
        self.order = []
        self.num_variables = num_variables
#         self.coverage = dict()

#     cpdef add_coverage(Basis self, int variable, MissingnessBasisFunction b1, \
#                        MissingnessBasisFunction b2):
#         cdef int index = len(self.order)
#         self.coverage[variable] = (index, index + 1)
#         self.append(b1)
#         self.append(b2)
#
#     cpdef get_coverage(Basis self, int variable):
#         cdef int idx1, idx2
#         idx1, idx2 = self.coverage[variable]
#         return self.order[idx1], self.order[idx2]
#
#     cpdef bint has_coverage(Basis self, int variable):
#         return variable in self.coverage

    def __reduce__(Basis self):
        return (self.__class__, (self.num_variables,), self._getstate())

    def _getstate(Basis self):
        return {'order': self.order}

    def __setstate__(Basis self, state):
        self.order = state['order']

    def __richcmp__(Basis self, other, method):
        if method == 2:
            return self._eq(other)
        elif method == 3:
            return not self._eq(other)
        else:
            return NotImplemented

    def _eq(Basis self, other):
        return (self.__class__ is other.__class__ and
                self._getstate() == other._getstate())

    def piter(Basis self):
        for bf in self.order:
            if not bf.is_pruned():
                yield bf

    def __str__(Basis self):
        cdef INDEX_t i
        cdef INDEX_t n = len(self)
        result = ''
        for i in range(n):
            result += str(self[i])
            result += '\n'
        return result

    cpdef int get_num_variables(Basis self):
        return self.num_variables

    cpdef dict anova_decomp(Basis self):
        '''
        See section 3.5, Friedman, 1991
        '''
        cdef INDEX_t bf_idx, n_bf = len(self)
        cdef dict result = {}
        cdef frozenset vars
        cdef BasisFunction bf
        for bf_idx in range(n_bf):
            bf = self.order[bf_idx]
            vars = frozenset(bf.variables())
            if vars in result:
                result[vars].append(bf)
            else:
                result[vars] = [bf]
        return result

    def smooth_knots(Basis self, mins, maxes):
        '''
        Used to find the side knots in the smoothed representation.
        '''
        cdef dict anova = self.anova_decomp()
        cdef dict intermediate = {}
        cdef dict result = {}
        for vars, bfs in anova.iteritems():
            intermediate[vars] = {}
            for var in vars:
                intermediate[vars][var] = []
            for bf in bfs:
                if bf.has_knot():
                    variable = bf.get_variable()
                    knot = bf.get_knot()
                    intermediate[vars][variable].append((bf, knot))
        for d in intermediate.itervalues():
            for var, lst in d.iteritems():
                lst.sort(key=lambda x: x[1])
                prev_minus = mins[var]
                prev = prev_minus
                prev_mid = prev_minus
                plus_idx = 0
                i = 0
                n_bfs = len(lst)
                while True:
                    if i >= n_bfs:
                        break
                    bf, knot = lst[i]
                    if knot > prev_mid:
                        prev = prev_mid
                    else:
                        prev = prev_minus
                    while plus_idx < n_bfs and lst[plus_idx][1] <= knot:
                        plus_idx += 1
                    if plus_idx < n_bfs and lst[plus_idx][1] > knot:
                        next = lst[plus_idx][1]
                    else:
                        next = maxes[var]
                    result[bf] = ((knot + prev) / 2.0, (knot + next) / 2.0)
                    prev_minus = prev
                    prev_mid = knot
                    i += 1
        return result

    cpdef smooth(Basis self, cnp.ndarray[FLOAT_t, ndim=2] X):
        mins = list(X.min(0))
        maxes = list(X.max(0))
        knot_dict = self.smooth_knots(mins, maxes)
        root = self[0]._get_root()
        translation_dict = {}
        root.smooth(knot_dict, translation_dict)
        new_order = [translation_dict[bf] for bf in self]
        result = Basis(self.num_variables)
        for bf in new_order:
            result.append(bf)
        return result

    cpdef append(Basis self, BasisFunction basis_function):
        self.order.append(basis_function)

    def __iter__(Basis self):
        return self.order.__iter__()

    def __len__(Basis self):
        return self.order.__len__()

    cpdef BasisFunction get(Basis self, INDEX_t i):
        return self.order[i]

    def __getitem__(Basis self, INDEX_t i):
        return self.get(i)

    cpdef INDEX_t plen(Basis self):
        cdef INDEX_t length = 0
        for bf in self.order:
            if not bf.is_pruned():
                length += 1
        return length

    cpdef transform(Basis self, cnp.ndarray[FLOAT_t, ndim=2] X,
                    cnp.ndarray[BOOL_t, ndim=2] missing,
                    cnp.ndarray[FLOAT_t, ndim=2] B):
        cdef BasisFunction bf
        cdef INDEX_t col = 0
        for bf in self.order:
            if not bf.is_pruned():
                bf.apply(X, missing, B[:, col], recurse=True)
                col += 1

    cpdef weighted_transform(Basis self, cnp.ndarray[FLOAT_t, ndim=2] X,
                             cnp.ndarray[BOOL_t, ndim=2] missing,
                             cnp.ndarray[FLOAT_t, ndim=2] B,
                             cnp.ndarray[FLOAT_t, ndim=1] weights):
        self.transform(X, missing, B)
        apply_weights_2d(B, weights)

    cpdef transform_deriv(Basis self, cnp.ndarray[FLOAT_t, ndim=2] X,
                          cnp.ndarray[BOOL_t, ndim=2] missing,
                          cnp.ndarray[FLOAT_t, ndim=1] b,
                          cnp.ndarray[FLOAT_t, ndim=1] j,
                          cnp.ndarray[FLOAT_t, ndim=2] coef,
                          cnp.ndarray[FLOAT_t, ndim=3] J,
                          list variables_of_interest, bool prezeroed_j=False):

        cdef BasisFunction bf
        cdef INDEX_t i, j_, m, n

        # Zero out J if necessary
        m = J.shape[0]
        n = J.shape[1]
        p = J.shape[2]
        if not prezeroed_j:
            for j_ in range(n):
                for i in range(m):
                    for p_ in range(p):
                        J[i, j_, p_] = 0.0
        cdef INDEX_t var, bf_idx, coef_idx, n_bfs = len(self)
        cdef set variables

        for p_ in range(p):
            # Compute the derivative for each variable
            for j_, var in enumerate(variables_of_interest):
                coef_idx=0
                for bf_idx in range(n_bfs):
                    bf = self.order[bf_idx]
                    variables = bf.variables()
                    if (variables and var not in variables) or bf.is_pruned():
                        continue
                    bf.apply_deriv(X, missing, b, j, var)
                    for i in range(m):
                        J[i, j_, p_] += coef[p_, coef_idx] * j[i]
                    coef_idx += 1



from cpython cimport bool
cimport numpy as cnp
from ._types cimport FLOAT_t, INT_t, INDEX_t, BOOL_t

cdef class BasisFunction:
    '''Abstract.  Subclasses must implement the apply and __init__ methods.'''

    cdef BasisFunction parent
    cdef dict child_map
    cdef list children
    cdef bint pruned
    cdef bint prunable
    cdef bint splittable

    cpdef smooth(BasisFunction self, dict knot_dict, dict translation)

    cpdef bint has_knot(BasisFunction self)

    cpdef bint is_prunable(BasisFunction self)

    cpdef bint is_pruned(BasisFunction self)

    cpdef bint is_splittable(BasisFunction self)

    cpdef bint make_splittable(BasisFunction self)

    cpdef bint make_unsplittable(BasisFunction self)

    cpdef list get_children(BasisFunction self)

    cpdef BasisFunction get_coverage(BasisFunction self, INDEX_t variable)

    cpdef bool has_linear(BasisFunction self, INDEX_t variable)

    cpdef bool linear_in(BasisFunction self, INDEX_t variable)

    cpdef _set_parent(BasisFunction self, BasisFunction parent)

    cpdef _add_child(BasisFunction self, BasisFunction child)

    cpdef BasisFunction get_parent(BasisFunction self)

    cpdef prune(BasisFunction self)

    cpdef unprune(BasisFunction self)

    cpdef knots(BasisFunction self, INDEX_t variable)

    cpdef INDEX_t effective_degree(BasisFunction self)

    cpdef apply(BasisFunction self, cnp.ndarray[FLOAT_t, ndim=2] X,
                cnp.ndarray[BOOL_t, ndim=2] missing,
                cnp.ndarray[FLOAT_t, ndim=1] b, bint recurse= ?)

    cpdef cnp.ndarray[INT_t, ndim = 1] valid_knots(BasisFunction self,
        cnp.ndarray[FLOAT_t, ndim=1] values,
        cnp.ndarray[FLOAT_t, ndim=1] variable,
        int variable_idx, INDEX_t check_every,
        int endspan, int minspan,
        FLOAT_t minspan_alpha, INDEX_t n,
        cnp.ndarray[INT_t, ndim=1] workspace)

cdef class RootBasisFunction(BasisFunction):
    cpdef bint covered(RootBasisFunction self, INDEX_t variable)

    cpdef bint eligible(RootBasisFunction self, INDEX_t variable)

    cpdef set variables(RootBasisFunction self)

    cpdef _smoothed_version(RootBasisFunction self, BasisFunction parent,
                            dict knot_dict, dict translation)

    cpdef INDEX_t degree(RootBasisFunction self)

    cpdef _effective_degree(RootBasisFunction self, dict data_dict, dict missing_dict)

    cpdef _set_parent(RootBasisFunction self, BasisFunction parent)

    cpdef BasisFunction get_parent(RootBasisFunction self)

    cpdef apply(RootBasisFunction self, cnp.ndarray[FLOAT_t, ndim=2] X,
                cnp.ndarray[BOOL_t, ndim=2] missing,
                cnp.ndarray[FLOAT_t, ndim=1] b, bint recurse=?)

    cpdef apply_deriv(RootBasisFunction self, cnp.ndarray[FLOAT_t, ndim=2] X,
                      cnp.ndarray[BOOL_t, ndim=2] missing,
                      cnp.ndarray[FLOAT_t, ndim=1] b,
                      cnp.ndarray[FLOAT_t, ndim=1] j, INDEX_t var)

cdef class ConstantBasisFunction(RootBasisFunction):

    cpdef inline FLOAT_t eval(ConstantBasisFunction self)

    cpdef inline FLOAT_t eval_deriv(ConstantBasisFunction self)

cdef class VariableBasisFunction(BasisFunction):
    cdef INDEX_t variable
    cdef readonly label

    cpdef INDEX_t degree(VariableBasisFunction self)

    cpdef set variables(VariableBasisFunction self)

    cpdef INDEX_t get_variable(VariableBasisFunction self)

cdef class DataVariableBasisFunction(VariableBasisFunction):
    cpdef _effective_degree(DataVariableBasisFunction self, dict data_dict, dict missing_dict)

    cpdef bint covered(DataVariableBasisFunction self, INDEX_t variable)

    cpdef bint eligible(DataVariableBasisFunction self, INDEX_t variable)

    cpdef apply(DataVariableBasisFunction self, cnp.ndarray[FLOAT_t, ndim=2] X,
                cnp.ndarray[BOOL_t, ndim=2] missing,
                cnp.ndarray[FLOAT_t, ndim=1] b, bint recurse=?)

    cpdef apply_deriv(DataVariableBasisFunction self,
                      cnp.ndarray[FLOAT_t, ndim=2] X,
                      cnp.ndarray[BOOL_t, ndim=2] missing,
                      cnp.ndarray[FLOAT_t, ndim=1] b,
                      cnp.ndarray[FLOAT_t, ndim=1] j, INDEX_t var)

cdef class MissingnessBasisFunction(VariableBasisFunction):
    cdef readonly bint complement

    cpdef _effective_degree(MissingnessBasisFunction self, dict data_dict, dict missing_dict)

    cpdef bint covered(MissingnessBasisFunction self, INDEX_t variable)

    cpdef bint eligible(MissingnessBasisFunction self, INDEX_t variable)

    cpdef bint covered(MissingnessBasisFunction self, INDEX_t variable)

    cpdef bint eligible(MissingnessBasisFunction self, INDEX_t variable)

    cpdef apply(MissingnessBasisFunction self, cnp.ndarray[FLOAT_t, ndim=2] X,
                cnp.ndarray[BOOL_t, ndim=2] missing,
                cnp.ndarray[FLOAT_t, ndim=1] b, bint recurse=?)

    cpdef apply_deriv(MissingnessBasisFunction self,
                      cnp.ndarray[FLOAT_t, ndim=2] X,
                      cnp.ndarray[BOOL_t, ndim=2] missing,
                      cnp.ndarray[FLOAT_t, ndim=1] b,
                      cnp.ndarray[FLOAT_t, ndim=1] j, INDEX_t var)

    cpdef _smoothed_version(MissingnessBasisFunction self, BasisFunction parent,
                            dict knot_dict, dict translation)

cdef class HingeBasisFunctionBase(DataVariableBasisFunction):
    cdef FLOAT_t knot
    cdef INDEX_t knot_idx
    cdef bint reverse

    cpdef bint has_knot(HingeBasisFunctionBase self)

    cpdef INDEX_t get_variable(HingeBasisFunctionBase self)

    cpdef FLOAT_t get_knot(HingeBasisFunctionBase self)

    cpdef bint get_reverse(HingeBasisFunctionBase self)

    cpdef INDEX_t get_knot_idx(HingeBasisFunctionBase self)

cdef class SmoothedHingeBasisFunction(HingeBasisFunctionBase):
    cdef FLOAT_t p
    cdef FLOAT_t r
    cdef FLOAT_t knot_minus
    cdef FLOAT_t knot_plus

    cpdef _smoothed_version(SmoothedHingeBasisFunction self,
                            BasisFunction parent, dict knot_dict,
                            dict translation)

    cpdef get_knot_minus(SmoothedHingeBasisFunction self)

    cpdef get_knot_plus(SmoothedHingeBasisFunction self)

    cpdef _init_p_r(SmoothedHingeBasisFunction self)

    cpdef get_p(SmoothedHingeBasisFunction self)

    cpdef get_r(SmoothedHingeBasisFunction self)

cdef class HingeBasisFunction(HingeBasisFunctionBase):

    cpdef _smoothed_version(HingeBasisFunction self,
                            BasisFunction parent,
                            dict knot_dict, dict translation)

cdef class LinearBasisFunction(DataVariableBasisFunction):
    cpdef bool linear_in(LinearBasisFunction self, INDEX_t variable)

    cpdef _smoothed_version(LinearBasisFunction self, BasisFunction parent,
                            dict knot_dict, dict translation)

cdef class Basis:
    '''A wrapper that provides functionality related to a set of BasisFunctions
    with a common RootBasisFunction ancestor.  Retains the order in which
    BasisFunctions are added.'''

    cdef list order
    cdef readonly INDEX_t num_variables
#     cdef dict coverage

#     cpdef add_coverage(Basis self, int variable, MissingnessBasisFunction b1, \
#                        MissingnessBasisFunction b2)
#
#     cpdef get_coverage(Basis self, int variable)
#
#     cpdef bint has_coverage(Basis self, int variable)

    cpdef int get_num_variables(Basis self)

    cpdef dict anova_decomp(Basis self)

    cpdef smooth(Basis self, cnp.ndarray[FLOAT_t, ndim=2] X)

    cpdef append(Basis self, BasisFunction basis_function)

    cpdef INDEX_t plen(Basis self)

    cpdef BasisFunction get(Basis self, INDEX_t i)

    cpdef transform(Basis self, cnp.ndarray[FLOAT_t, ndim=2] X,
                    cnp.ndarray[BOOL_t, ndim=2] missing,
                    cnp.ndarray[FLOAT_t, ndim=2] B)

    cpdef weighted_transform(Basis self, cnp.ndarray[FLOAT_t, ndim=2] X,
                             cnp.ndarray[BOOL_t, ndim=2] missing,
                             cnp.ndarray[FLOAT_t, ndim=2] B,
                             cnp.ndarray[FLOAT_t, ndim=1] weights)

    cpdef transform_deriv(Basis self, cnp.ndarray[FLOAT_t, ndim=2] X,
                          cnp.ndarray[BOOL_t, ndim=2] missing,
                          cnp.ndarray[FLOAT_t, ndim=1] b,
                          cnp.ndarray[FLOAT_t, ndim=1] j,
                          cnp.ndarray[FLOAT_t, ndim=2] coef,
                          cnp.ndarray[FLOAT_t, ndim=3] J,
                          list variables_of_interest, bool prezeroed_j=?)
